/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999, 2000 Rainer Wichmann                                */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>


#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)

#include "sh_hash.h"
#include "sh_utils.h"
#include "sh_error.h"
#include "sh_tiger.h"

#if defined(SH_WITH_CLIENT)
#include "sh_forward.h"
#endif

#undef  FIL__
#define FIL__  _("sh_hash.c")

/* MAX_PATH_STORE must be >= KEY_LEN
 */
#define MAX_PATH_STORE SH_BUFSIZE+1

typedef struct store_info {
  unsigned short   mark;
  /*
  char             fullpath[MAX_PATH_STORE+2];
  char             linkpath[MAX_PATH_STORE+2];
  */
  unsigned int     mode;
  unsigned int     linkmode;
  char             c_mode[11];
#if defined(__linux__)
  unsigned long    attributes;
  char             c_attributes[16];
#endif
  long             owner;
  char             c_owner[USER_MAX+2];
  long             group;
  char             c_group[GROUP_MAX+2];
  unsigned long    hardlinks;
  unsigned long    ino;
  unsigned long    size;
  unsigned long    atime;
  unsigned long    mtime;
  unsigned long    ctime;
  char             checksum[KEY_LEN+1];
} sh_filestore_t;
  
typedef struct file_info {
  sh_filestore_t   theFile;
  char           * fullpath;
  char           * linkpath;
  int              visited;
  int              reported;
  struct           file_info * next;
} sh_file_t;

  static char  *policy[] = {
    N_("[]"),
    N_("[ReadOnly]"),
    N_("[LogFiles]"),
    N_("[GrowingLogs]"),
    N_("[IgnoreNone]"),
    N_("[IgnoreAll]"),
    N_("[Attributes]"),
    NULL
  };


/**********************************
 *
 * hash table functions
 *
 **********************************
 */

#include "sh_hash.h"

/* must fit an int              */
#define TABSIZE 2048  

/* must fit an unsigned short   */
/* changed for V0.8, as the     */
/* database format has changed  */

/* changed again for V0.9       */
#define REC_MAGIC 19

static sh_file_t * tab[TABSIZE];

/**************************************************************
 *
 * compute hash function
 *
 **************************************************************/
static int hashfunc(char *s) 
{
  unsigned n = 0; 

  for ( ; *s; s++) 
    n = 31 * n + *s; 
  return n % TABSIZE; 
} 


/**************************************************************
 *
 * search for files not visited, and check whether they exist
 *
 **************************************************************/
static void hash_unvisited (sh_file_t *p, ShErrLevel level)
{
  struct stat buf;
  int i;
  char * tmp;

  if (p->next != NULL)
    hash_unvisited (p->next, level);

  if (p->visited == S_FALSE && p->reported == S_FALSE)
    {
      sl_set_suid(); 
      i = retry_lstat(p->fullpath, &buf);
      sl_unset_suid();

      if (0 != i)
	{
	  tmp = sh_util_safe_name(p->fullpath);
	  sh_error_handle (level, FIL__, __LINE__, 0, MSG_FI_MISS, tmp);
	  SH_FREE(tmp);
	  if (sh.flag.reportonce == S_TRUE)
	    p->reported = S_TRUE;
	}
    }

  if (sh.flag.reportonce == S_FALSE)
    p->reported = S_FALSE;
  p->visited = S_FALSE;
  return;
}


/*********************************************************************
 *
 * Search for files in the database that have been deleted from disk.
 *
 *********************************************************************/
void sh_hash_unvisited (ShErrLevel level)
{
  int i;

  for (i = 0; i < TABSIZE; ++i)
    {
      if (tab[i] != NULL) 
	hash_unvisited (tab[i], level);
    }
  return;
}


/**********************************************************************
 *
 * delete hash array
 *
 **********************************************************************/
static void hash_kill (sh_file_t *p)
{

  if (p->next != NULL)
    hash_kill (p->next);

  SH_FREE(p->fullpath);
  SH_FREE(p->linkpath);
  SH_FREE(p);
  return;
}


/***********************************************************************
 *
 * get info out of hash array
 *
 ***********************************************************************/
static sh_file_t * hashsearch (char * s) 
{
  sh_file_t * p;

  for (p = tab[hashfunc(s)]; p; p = p->next) 
    if (sl_strcmp(s, p->fullpath) == 0) 
      return p; 
  return NULL;
} 


/***********************************************************************
 *
 * insert into hash array
 *
 ***********************************************************************/
static void hashinsert (sh_file_t * s) 
{
  sh_file_t * p;

  int key = hashfunc(s->fullpath);

  if (tab[key] == NULL) 
    {
      tab[key] = s;
      tab[key]->next = NULL;
      return;
    } 
  else 
    {
      p = tab[key];
      while (1) 
	{
	  if (p->next == NULL) 
	    {
	      p->next = s;
	      p->next->next = NULL;
	      return;
	    }
	  p = p->next;
	}
    }
  /* notreached */
}


/******************************************************************
 *
 * ------- Check functions -------
 *
 ******************************************************************/

static int IsInit = 0;

/******************************************************************
 *
 * Initialize
 *
 ******************************************************************/
void sh_hash_init ()
{
  sh_file_t * p;
  sh_filestore_t ft;
  static SL_TICKET fd = (-1);
  long i;
  int count = 0;
  char * line = NULL;

  char * fullpath = NULL;
  char * linkpath = NULL;

  if (IsInit == 1) 
    return;

  for (i = 0; i < TABSIZE; ++i) 
    tab[i] = NULL;

#if defined(SH_WITH_CLIENT)
  if (fd == (-1) && 0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
    {
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_D_DSTART);
      fd = sh_forward_req_file(_("DATA"));
      if (SL_ISERROR(fd))
	exit (EXIT_FAILURE);

      sl_rewind (fd);

      tiger_fd = fd;
      sl_strlcpy (sh.data.hash, 
		  sh_tiger_hash (file_path('C', 'R'), 
				 TIGER_FD, 0),
		  KEY_LEN+1);
    }
  else 
#endif
    if (fd == (-1))
      {
	if ( SL_ISERROR(fd = sl_open_read(file_path('D', 'R'), SL_YESPRIV))) 
	  {
	    sh_error_handle ((-1), FIL__, __LINE__, fd, MSG_EXIT_ABORT1, 
			     sh.prg_name);
	    aud_exit (FIL__, __LINE__, EXIT_FAILURE);
	  }
	
	tiger_fd = fd;
	if (0 != sl_strncmp(sh.data.hash, 
			    sh_tiger_hash (file_path('D', 'R'), TIGER_FD, 0),
			    KEY_LEN)
	    && sh.flag.checkSum != SH_CHECK_INIT) 
	  {
	    sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_E_AUTH,
			     file_path('D', 'R'));
	    aud_exit (FIL__, __LINE__, EXIT_FAILURE);
	  }
      }


  sl_rewind (fd);

  /* fast forward to start of data
   */
  line = SH_ALLOC(SH_BUFSIZE+1);

  while (1) 
    {
      i =  sh_unix_getline (fd, line, SH_BUFSIZE);
      if (i < 0 ) 
	{
	  SH_FREE(line);
	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_P_NODATA,
			   file_path('D', 'R'));
	  aud_exit (FIL__, __LINE__, EXIT_FAILURE);
	}

#if defined(SH_STEALTH)
    if (0 == sl_strncmp (line, N_("[SOF]"), 5)) 
#else
    if (0 == sl_strncmp (line, _("[SOF]"),  5)) 
#endif
      break;
    }


  while (1) 
    {
      p = SH_ALLOC(sizeof(sh_file_t));
      ++count;
      i = sl_read(fd, &ft, sizeof(sh_filestore_t));
      if ( SL_ISERROR(i) || i == 0) 
	{
	  SH_FREE(line);
	  line = NULL;
	  SH_FREE(p);
	  break;
	}
      
      if (ft.mark != REC_MAGIC)
	{
	  SH_FREE(line);
	  line = NULL;
	  SH_FREE(p);
	  break;
	}

      i =  sh_unix_getline (fd, line, SH_BUFSIZE);
      if (i < 0 ) 
	{
	  SH_FREE(line);
	  SH_FREE(p);
	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_P_NODATA,
			   file_path('D', 'R'));
	  aud_exit (FIL__, __LINE__,EXIT_FAILURE);
	}

      i = sl_strlen(line)+1;
      fullpath = SH_ALLOC(i);
      sl_strlcpy (fullpath, line, i);
      if (fullpath[i-2] == '\n')
	fullpath[i-2] = '\0';

      i =  sh_unix_getline (fd, line, SH_BUFSIZE);
      if (i < 0 ) 
	{
	  SH_FREE(line);
	  SH_FREE(fullpath);
	  SH_FREE(p);
	  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_P_NODATA,
			   file_path('D', 'R'));
	  aud_exit (FIL__, __LINE__,EXIT_FAILURE);
	}

      i = sl_strlen(line)+1;
      linkpath = SH_ALLOC(i);
      sl_strlcpy (linkpath, line, i);
      if (linkpath[i-2] == '\n')
	linkpath[i-2] = '\0';

#if defined(SH_STEALTH)
      sh_do_decode(fullpath,    sl_strlen(fullpath));

#if defined(__linux__)
      sh_do_decode(ft.c_attributes,   sl_strlen(ft.c_attributes));
#endif

      sh_do_decode(ft.c_mode,   sl_strlen(ft.c_mode));
      sh_do_decode(ft.c_owner,  sl_strlen(ft.c_owner));
      sh_do_decode(ft.c_group,  sl_strlen(ft.c_group));
      sh_do_decode(ft.checksum, sl_strlen(ft.checksum));


      if (ft.c_mode[0] == 'l') 
	{  
	  sh_do_decode(linkpath, sl_strlen(linkpath));
	}
#endif

      memcpy( &(*p).theFile, &ft, sizeof(sh_filestore_t) );
      p->visited  = S_FALSE;
      p->reported = S_FALSE;
      p->fullpath = fullpath;
      p->linkpath = linkpath;
      
      hashinsert (p); 
      ft.mark = (REC_MAGIC + 5);
    }

  if (line != NULL)
    SH_FREE(line);

#if defined(SH_WITH_CLIENT)
  /****************************
  if (0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
    sl_rewind (fd);
  else
    {
      sl_close (fd);
      fd = -1;
    }
  **************************/

  /* keep db in memory, so we have no open file
   */
  sl_close (fd);
  fd = -1;

#else
  sl_close (fd);
  fd = -1;
#endif

  /* --- Initialization done successfully. ---
   */
  IsInit = 1;
  return;
}
  
/*****************************************************************
 *
 * delete hash array
 *
 *****************************************************************/
void sh_hash_hashdelete ()
{
  int i;

  if (IsInit == 0) 
    return;
  for (i = 0; i < TABSIZE; ++i) 
    if (tab[i] != NULL) hash_kill (tab[i]);
  IsInit = 0;
  return;
}

/******************************************************************
 *
 * Insert a file into the database.
 *
 ******************************************************************/ 
void sh_hash_pushdata (file_type * buf, char * fileHash)
{
  SL_TICKET fd;
  sh_filestore_t p;
  static int isfirst = 1;

  struct stat sbuf;

  char fullpath[MAX_PATH_STORE+1];
  char linkpath[MAX_PATH_STORE+1];

  linkpath[0] =  '-'; 
  linkpath[1] = '\0'; 
  fullpath[0] =  '-'; 
  fullpath[1] = '\0'; 

  if (0 == sl_strcmp(file_path('D', 'W'), _("REQ_FROM_SERVER")))
    return;

  if (isfirst == 1)
    {
      if (0 == retry_lstat(file_path('D', 'W'), &sbuf))
	sh_error_handle((-1), FIL__, __LINE__, 0, MSG_FI_DBEX,
			file_path('D', 'W'));
    }

  if ( SL_ISERROR(fd = sl_open_write(file_path('D', 'W'), SL_YESPRIV)))
    return;

  if ( SL_ISERROR(sl_forward(fd)))
    return;

  if (sl_strlen(buf->fullpath) <= MAX_PATH_STORE) 
    {
      sl_strlcpy(fullpath, buf->fullpath, MAX_PATH_STORE+1);
      if (sl_strlen(buf->fullpath) < (MAX_PATH_STORE-3))
	{
	  fullpath[MAX_PATH_STORE-2] = '\0';
	  fullpath[MAX_PATH_STORE-1] = '\n';
	}
    } 
  else 
    {
      sl_strlcpy(fullpath, 
		 sh_tiger_hash (buf->fullpath,
				TIGER_DATA,sl_strlen(buf->fullpath)), 
		 KEY_LEN+1);
    }

#if defined(SH_STEALTH)
  sh_do_encode(fullpath, sl_strlen(fullpath));
#endif

  if (buf->c_mode[0] == 'l') 
    {  
      if (sl_strlen(buf->linkpath) <= MAX_PATH_STORE) 
	{
	  sl_strlcpy(linkpath, buf->linkpath, MAX_PATH_STORE+1);  
	} 
      else 
	{
	  sl_strlcpy(linkpath, 
		     sh_tiger_hash (buf->linkpath,
				    TIGER_DATA,sl_strlen(buf->linkpath)),
		     KEY_LEN+1);
	}

#if defined(SH_STEALTH)
      sh_do_encode(linkpath, sl_strlen(linkpath));
#endif
    }

  p.mark = REC_MAGIC;
  sl_strlcpy(p.c_mode,   buf->c_mode,   11);
  sl_strlcpy(p.c_group,  buf->c_group,  GROUP_MAX+1);
  sl_strlcpy(p.c_owner,  buf->c_owner,  USER_MAX+1);
  sl_strlcpy(p.checksum, fileHash,      KEY_LEN+1);
#if defined(__linux__)
  sl_strlcpy(p.c_attributes, buf->c_attributes, 13);
#endif

#if defined(SH_STEALTH)
  sh_do_encode(p.c_mode,   sl_strlen(p.c_mode));
  sh_do_encode(p.c_owner,  sl_strlen(p.c_owner));
  sh_do_encode(p.c_group,  sl_strlen(p.c_group));
  sh_do_encode(p.checksum, sl_strlen(p.checksum));
#if defined(__linux__)
  sh_do_encode(p.c_attributes,   sl_strlen(p.c_attributes));
#endif
#endif

#if defined(__linux__)
  p.attributes  = buf->attributes;
#endif
  p.linkmode    = buf->linkmode;
  p.hardlinks   = buf->hardlinks;
  p.mode  = buf->mode;
  p.ino   = buf->ino;
  p.size  = buf->size;
  p.mtime = buf->mtime;
  p.atime = buf->atime;
  p.ctime = buf->ctime;
  p.owner = buf->owner;
  p.group = buf->group;

  /* write the start marker 
   */
  if (isfirst == 1) 
    {
#if defined(SH_STEALTH)
      sl_write      (fd,        "\n", 1);
      sl_write_line (fd, N_("[SOF]"), 5);
#else
      sl_write_line (fd, _("\n[SOF]"),  6);
#endif
      isfirst = 0;
    }
      
  sl_write      (fd,       &p, sizeof(sh_filestore_t));
  sl_write_line (fd, fullpath,    sl_strlen(fullpath));
  sl_write_line (fd, linkpath,    sl_strlen(linkpath));

  sl_close (fd);
  return;
}

/*********************************************************************
 *
 * Check whether a file is present in the database.
 *
 *********************************************************************/
int sh_hash_have_it (char * newname)
{
  sh_file_t * p;

  if (newname == NULL)
    return (-1);

  if (IsInit != 1) 
    sh_hash_init();
  if (sl_strlen(newname) <= MAX_PATH_STORE) 
    p = hashsearch(newname);
  else 
    p = hashsearch ( sh_tiger_hash(newname, TIGER_DATA, sl_strlen(newname)) );
  if (p == NULL) 
    return (-1);
  return (0);
}

/*****************************************************************
 *
 * Set a file's status to 'visited'. This is required for
 * files that should be ignored, and may be present in the
 * database, but not on disk.
 *
 *****************************************************************/
int sh_hash_set_visited (char * newname)
{
  sh_file_t * p;

  if (newname == NULL)
    return (-1);
  if (IsInit != 1) 
    sh_hash_init();

  if (sl_strlen(newname) <= MAX_PATH_STORE) 
    p = hashsearch(newname);
  else 
    p = hashsearch (sh_tiger_hash(newname, TIGER_DATA, sl_strlen(newname)));
  
  if (p == NULL) 
    return (-1);
  p->visited  = S_TRUE;
  p->reported = S_TRUE;

  return 0;
}



/*****************************************************************
 *
 * Compare a file with the database status.
 *
 *****************************************************************/
void sh_hash_compdata (int class, file_type * theFile, char * fileHash)
{
  char * msg;
  sh_file_t * p;
  char * tmp;
  char * tmp_path;
  char * tmp_lnk;
  char * tmp_lnk_old;
  char timstr1c[32];
  char timstr2c[32];
  char timstr1a[32];
  char timstr2a[32];
  char timstr1m[32];
  char timstr2m[32];
  char linkHash[KEY_LEN+1];
  int  maxcomp;

  unsigned long modi_mask = 0;

  if (IsInit != 1) sh_hash_init();

  /* --------  find the entry for the file ----------------       */

  if (sl_strlen(theFile->fullpath) <= MAX_PATH_STORE) 
    p = hashsearch(theFile->fullpath);
  else 
    p = hashsearch( sh_tiger_hash(theFile->fullpath, 
				  TIGER_DATA, 
				  sl_strlen(theFile->fullpath))
		    );


  /* --------- Not found. ------------
   */

  if (p == NULL) 
    {
      if (theFile->reported == S_FALSE)
	{
	  tmp = sh_util_safe_name(theFile->fullpath);
	  sh_error_handle (ShDFLevel[class], FIL__, __LINE__, 0, MSG_FI_ADD, 
			   tmp);
	  SH_FREE(tmp);
	  if (sh.flag.reportonce == S_TRUE)
	    theFile->reported = S_TRUE;
	}
      return;
    }

  TPT ((0, FIL__, __LINE__, _("file=<%s>, cs_old=<%s>, cs_new=<%s>\n"),
	theFile->fullpath, fileHash, p->theFile.checksum));

  if ( sl_strncmp (fileHash, p->theFile.checksum, 50) != 0 && 
       (theFile->check_mask & MODI_CHK) != 0)
    {
      modi_mask |= MODI_CHK;
      TPT ((0, FIL__, __LINE__, _("mod=<checksum>")));
    } 

  if (p->theFile.c_mode[0] == 'l') 
    {
      if (sl_strlen(theFile->linkpath) >= MAX_PATH_STORE) 
	{
	  sl_strlcpy(linkHash, 
		     sh_tiger_hash(theFile->linkpath, 
				   TIGER_DATA,
				   sl_strlen(theFile->linkpath)), 
		     MAX_PATH_STORE+1);
	  maxcomp = MAX_PATH_STORE;
	} 
      else 
	{
	  sl_strlcpy(linkHash, theFile->linkpath, KEY_LEN + 1);
	  maxcomp = KEY_LEN;
	}

    if ( sl_strncmp (linkHash, p->linkpath, maxcomp) != 0 &&
	 (theFile->check_mask & MODI_LNK) != 0)
      {
	modi_mask |= MODI_LNK;
	TPT ((0, FIL__, __LINE__, _("mod=<link>")));
      } 
    }
    
  if ( theFile->ino != p->theFile.ino  &&
       (theFile->check_mask & MODI_INO) != 0)
    {
      modi_mask |= MODI_INO;
      TPT ((0, FIL__, __LINE__, _("mod=<inode>")));
    } 
    
  if ( theFile->hardlinks != p->theFile.hardlinks &&
       (theFile->check_mask & MODI_HLN) != 0)
    {
      modi_mask |= MODI_HLN;
      TPT ((0, FIL__, __LINE__, _("mod=<hardlink>")));
    } 
    
  if ( (  (theFile->mode != p->theFile.mode)
#if defined(__linux__)
	  || (theFile->attributes != p->theFile.attributes)
#endif
	)
       && (theFile->check_mask & MODI_MOD) != 0)
    {
      modi_mask |= MODI_MOD;
      TPT ((0, FIL__, __LINE__, _("mod=<mode>")));
    } 

  if ( theFile->owner != p->theFile.owner &&
       (theFile->check_mask & MODI_USR) != 0)
    {
      modi_mask |= MODI_USR;
      TPT ((0, FIL__, __LINE__, _("mod=<user>")));
    } 

  if ( theFile->group != p->theFile.group &&
       (theFile->check_mask & MODI_GRP) != 0)
    {
      modi_mask |= MODI_GRP;
      TPT ((0, FIL__, __LINE__, _("mod=<group>")));
    } 

  
  if ( theFile->mtime != p->theFile.mtime &&
       (theFile->check_mask & MODI_MTM) != 0)
    {
      modi_mask |= MODI_MTM;
      TPT ((0, FIL__, __LINE__, _("mod=<mtime>")));
    } 
  
  if ( theFile->atime != p->theFile.atime &&
       (theFile->check_mask & MODI_ATM) != 0)
    {
      modi_mask |= MODI_ATM;
      TPT ((0, FIL__, __LINE__, _("mod=<atime>")));
    } 

  
  /* Resetting the access time will set a new ctime. Thus, either we ignore
   * the access time or the ctime for NOIGNORE
   */
  if ( theFile->ctime != p->theFile.ctime &&
       (theFile->check_mask & MODI_CTM) != 0)
    {
      modi_mask |= MODI_CTM;
      TPT ((0, FIL__, __LINE__, _("mod=<ctime>")));
    } 

  if ( theFile->size != p->theFile.size &&
       (theFile->check_mask & MODI_SIZ) != 0)
    {
      if (class == SH_LEVEL_LOGGROW && theFile->size < p->theFile.size)
	{
	  modi_mask |= MODI_SIZ;
	  TPT ((0, FIL__, __LINE__, _("mod=<size>")));
	} 
      else if (class != SH_LEVEL_LOGGROW)
	{ 
	  modi_mask |= MODI_SIZ;
	  TPT ((0, FIL__, __LINE__, _("mod=<size>")));
	} 
    }

  /* --- Report full details. ---
   */
  if (modi_mask != 0 && sh.flag.fulldetail == S_TRUE)
    {
      if ((theFile->check_mask & MODI_ATM) == 0)
	modi_mask = MASK_READONLY;
      else
	modi_mask = MASK_NOIGNORE;
    }

  /* --- Report on modified files. ---
   */
  if (modi_mask != 0 && p->reported == S_FALSE)
    { 
      tmp = SH_ALLOC(SH_BUFSIZE);
      msg = SH_ALLOC(SH_BUFSIZE);
      msg[0] = '\0';

      if ((modi_mask & MODI_MOD) != 0)
	{
#if defined(__linux__)
	  sl_snprintf(tmp, SH_BUFSIZE, 
		      _("mode_old=<%s>, mode_new=<%s>, attr_old=<%s>, attr_new=<%s>, "),
		      p->theFile.c_mode, theFile->c_mode,
		      p->theFile.c_attributes, theFile->c_attributes);
#else
	  sl_snprintf(tmp, SH_BUFSIZE, _("mode_old=<%s>, mode_new=<%s>, "),
		      p->theFile.c_mode, theFile->c_mode);
#endif
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_HLN) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, 
		      _("hardlinks_old=<%ld>, hardlinks_new=<%ld>, "),
		      p->theFile.hardlinks, theFile->hardlinks);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_INO) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, 
		      _("inode_old=<%ld>, inode_new=<%ld>, "),
		      p->theFile.ino, theFile->ino);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_USR) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, _("owner_old=<%s>, owner_new=<%s>, "),
		      p->theFile.c_owner, theFile->c_owner);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_GRP) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, _("group_old=<%s>, group_new=<%s>, "),
		      p->theFile.c_group, theFile->c_group);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_SIZ) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, _("size_old=<%ld>, size_new=<%ld>, "),
		      p->theFile.size, theFile->size);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_CTM) != 0)
	{
	  sl_strlcpy (timstr1c, sh_unix_time (p->theFile.ctime), 32);
	  sl_strlcpy (timstr2c, sh_unix_time (theFile->ctime),   32);
	  sl_snprintf(tmp, SH_BUFSIZE, _("ctime_old=<%s>, ctime_new=<%s>, "),
		      timstr1c, timstr2c);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_ATM) != 0)
	{
	  sl_strlcpy (timstr1a, sh_unix_time (p->theFile.atime), 32);
	  sl_strlcpy (timstr2a, sh_unix_time (theFile->atime),   32);
	  sl_snprintf(tmp, SH_BUFSIZE, _("atime_old=<%s>, atime_new=<%s>, "),
		      timstr1a, timstr2a);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}

      if ((modi_mask & MODI_MTM) != 0)
	{
	  sl_strlcpy (timstr1m, sh_unix_time (p->theFile.mtime), 32);
	  sl_strlcpy (timstr2m, sh_unix_time (theFile->mtime),   32);
	  sl_snprintf(tmp, SH_BUFSIZE, _("mtime_old=<%s>, mtime_new=<%s>, "),
		      timstr1m, timstr2m);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}


      if ((modi_mask & MODI_CHK) != 0)
	{
	  sl_snprintf(tmp, SH_BUFSIZE, 
		      _("chksum_old=<%s>, chksum_new=<%s>, "),
		      p->theFile.checksum, fileHash);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}


      if ((modi_mask & MODI_LNK) != 0 && theFile->c_mode[0] == 'l')
	{
	  tmp_lnk     = sh_util_safe_name(theFile->linkpath);
	  tmp_lnk_old = sh_util_safe_name(p->linkpath);
	  sl_snprintf(tmp, SH_BUFSIZE, _("link_old=<%s>, link_new=<%s>"),
		      tmp_lnk_old, tmp_lnk);
	  SH_FREE(tmp_lnk);
	  SH_FREE(tmp_lnk_old);
	  sl_strlcat(msg, tmp, SH_BUFSIZE); 
	}


      tmp_path = sh_util_safe_name(theFile->fullpath);
      sh_error_handle (ShDFLevel[class], FIL__, __LINE__, 0, MSG_FI_CHAN,
		       policy[class], tmp_path, msg);
      SH_FREE(tmp_path);
      SH_FREE(tmp);
      SH_FREE(msg);
      p->reported = S_TRUE;

    }

  p->visited = S_TRUE;

  return;
}

/* if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE) */

#endif
